import asyncio
import logging
from asyncio import iscoroutinefunction
from typing import Any, Callable, NamedTuple, Set, Tuple, Type

from croniter import croniter

from aiomisc import Service
from aiomisc.cron import CronCallback


log = logging.getLogger(__name__)
ExceptionsType = Tuple[Type[Exception], ...]


class StoreItem(NamedTuple):
    callback: CronCallback
    spec: str
    shield: bool
    suppress_exceptions: ExceptionsType


class CronService(Service):
    _callbacks_storage: Set[StoreItem]

    def __init__(self, **kwargs: Any):
        super(CronService, self).__init__(**kwargs)
        self._callbacks_storage = set()

    def register(
        self,
        function: Callable,
        spec: str,
        shield: bool = False,
        suppress_exceptions: ExceptionsType = (),
    ) -> None:
        if not iscoroutinefunction(function):
            raise TypeError("function should be a coroutine %r" % function)
        if not croniter.is_valid(spec):
            raise TypeError("Not valid cron spec %r" % spec)

        self._callbacks_storage.add(
            StoreItem(
                CronCallback(function), spec, shield, suppress_exceptions,
            ),
        )

    async def start(self) -> None:
        for item in self._callbacks_storage:
            item.callback.start(
                spec=item.spec,
                loop=self.loop,
                shield=item.shield,
                suppress_exceptions=item.suppress_exceptions,
            )
        log.info("Cron service %s started", self)

    async def stop(self, exception: Exception = None) -> None:
        async def _shutdown(item: StoreItem) -> None:
            await asyncio.gather(item.callback.stop(), return_exceptions=True)

        await asyncio.gather(
            *[_shutdown(store) for store in self._callbacks_storage],
            return_exceptions=True,
        )
        log.info("Cron service %s is stopped", self)



import asyncio
import inspect
import logging
import socket
import sys
from asyncio import Queue, ensure_future
from types import MappingProxyType
from typing import Any, Iterable, Mapping

import yarl
from aiohttp import ClientSession, TCPConnector
from raven import Client  # type: ignore
from raven.handlers.logging import SentryHandler  # type: ignore
from raven.transport import Transport  # type: ignore
from raven_aiohttp import QueuedAioHttpTransport  # type: ignore

from aiomisc.service import Service


log = logging.getLogger(__name__)


class DummyTransport(Transport):         # type: ignore
    def send(self, url, data, headers):  # type: ignore
        pass


class QueuedPatchedAioHttpTransport(QueuedAioHttpTransport):  # type: ignore

    def __init__(
        self,
        *args: Any,
        workers: int = 1,
        qsize: int = 1000,
        **kwargs: Any
    ):
        super(QueuedAioHttpTransport, self).__init__(*args, **kwargs)
        loop_args = (
            {"loop": self._loop} if sys.version_info < (3, 8) else {}
        )
        self._queue: Queue = Queue(maxsize=qsize, **loop_args)

        self._workers = set()

        for _ in range(workers):
            worker = ensure_future(self._worker())
            self._workers.add(worker)
            worker.add_done_callback(self._workers.remove)


class QueuedKeepaliveAioHttpTransport(
    QueuedPatchedAioHttpTransport,
):
    DNS_CACHE_TTL = 600
    DNS_CACHE = True
    TCP_CONNECTION_LIMIT = 32
    TCP_CONNECTION_LIMIT_HOST = 8
    WORKERS = 1
    QUEUE_SIZE = 1000

    def __init__(
        self, *args: Any, family: int = socket.AF_UNSPEC,
        loop: asyncio.AbstractEventLoop = None,
        dns_cache: bool = DNS_CACHE,
        dns_cache_ttl: int = DNS_CACHE_TTL,
        connection_limit: int = TCP_CONNECTION_LIMIT,
        connection_limit_host: int = TCP_CONNECTION_LIMIT_HOST,
        workers: int = WORKERS, qsize: int = QUEUE_SIZE, **kwargs: Any
    ):
        self.connection_limit = connection_limit
        self.connection_limit_host = connection_limit_host
        self.dns_cache = dns_cache
        self.dns_cache_ttl = dns_cache_ttl

        super().__init__(
            *args, family=family, loop=loop, keepalive=True,
            workers=workers, qsize=qsize, **kwargs
        )

    def _client_session_factory(self) -> ClientSession:
        self.connector = TCPConnector(
            family=self.family,
            limit=self.connection_limit,
            limit_per_host=self.connection_limit_host,
            ttl_dns_cache=self.dns_cache_ttl,
            use_dns_cache=self.dns_cache,
            verify_ssl=self.verify_ssl,
        )

        return ClientSession(
            connector=self.connector,
            connector_owner=False,
        )

    async def _close(self) -> Transport:
        transport = await super()._close()
        if inspect.iscoroutinefunction(self.connector.close()):
            await self.connection.close()
        else:
            self.connector.close()
        return transport


class RavenSender(Service):
    __required__ = "sentry_dsn",

    sentry_dsn: yarl.URL
    min_level: int = logging.WARNING
    client_options: Mapping[str, Any] = MappingProxyType({})
    client: Client
    filters: Iterable[logging.Filter] = ()

    async def start(self) -> None:
        self.client = Client(
            str(self.sentry_dsn),
            transport=QueuedKeepaliveAioHttpTransport,
            **self.client_options
        )

        # Initialize Transport object
        self.client.remote.get_transport()

        self.sentry_dsn = yarl.URL(
            self.sentry_dsn,
        ).with_password("").with_user("")

        log.info("Starting Raven for %r", self.sentry_dsn)

        handler = SentryHandler(
            client=self.client,
            level=self.min_level,
        )

        # Add filters
        for fltr in self.filters:
            handler.addFilter(fltr)

        logging.getLogger().handlers.append(
            handler,
        )

    async def stop(self, *_: Any) -> None:
        transport = self.client.remote.get_transport()
        self.client.set_dsn("", transport=DummyTransport)

        await transport.close()

        log.info("Stopping Raven client for %r", self.sentry_dsn)


__all__ = ("RavenSender",)



import asyncio
import contextvars
import inspect
import logging
import threading
import time
import warnings
from concurrent.futures import ThreadPoolExecutor as ThreadPoolExecutorBase
from functools import partial, wraps
from multiprocessing import cpu_count
from queue import SimpleQueue
from types import MappingProxyType
from typing import (
    Any, Awaitable, Callable, Coroutine, Dict, FrozenSet, NamedTuple, Optional,
    Set, Tuple, TypeVar,
)

from .compat import get_current_loop
from .counters import Statistic
from .iterator_wrapper import IteratorWrapper


T = TypeVar("T")
F = TypeVar("F", bound=Callable[..., Any])
log = logging.getLogger(__name__)


def context_partial(
    func: F, *args: Any,
    **kwargs: Any
) -> Any:
    context = contextvars.copy_context()
    return partial(context.run, func, *args, **kwargs)


class ThreadPoolException(RuntimeError):
    pass


class WorkItemBase(NamedTuple):
    func: Callable[..., Any]
    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]
    future: asyncio.Future
    loop: asyncio.AbstractEventLoop


class ThreadPoolStatistic(Statistic):
    threads: int
    done: int
    error: int
    success: int
    submitted: int
    sum_time: float


class WorkItem(WorkItemBase):
    @staticmethod
    def set_result(
        future: asyncio.Future, result: Any, exception: Exception,
    ) -> None:
        if future.done():
            return

        if exception:
            future.set_exception(exception)
        else:
            future.set_result(result)

    def __call__(self, statistic: ThreadPoolStatistic) -> None:
        if self.future.done():
            return

        result, exception = None, None

        delta = -self.loop.time()
        try:
            result = self.func(*self.args, **self.kwargs)
            statistic.success += 1
        except BaseException as e:
            statistic.error += 1
            exception = e
        finally:
            delta += time.monotonic()
            statistic.sum_time += delta
            statistic.done += 1

        if self.loop.is_closed():
            raise asyncio.CancelledError

        self.loop.call_soon_threadsafe(
            self.__class__.set_result,
            self.future,
            result,
            exception,
        )


class ThreadPoolExecutor(ThreadPoolExecutorBase):
    __slots__ = (
        "__futures", "__pool", "__tasks",
        "__write_lock", "__thread_events",
    )

    def __init__(
        self, max_workers: int = max((cpu_count(), 4)),
        loop: asyncio.AbstractEventLoop = None,
        statistic_name: Optional[str] = None,
    ) -> None:
        """"""
        if loop:
            warnings.warn(DeprecationWarning("loop argument is obsolete"))

        self.__futures: Set[asyncio.Future[Any]] = set()

        self.__thread_events: Set[threading.Event] = set()
        self.__tasks: SimpleQueue[Optional[WorkItem]] = SimpleQueue()
        self.__write_lock = threading.RLock()
        self._statistic = ThreadPoolStatistic(statistic_name)

        pools = set()
        for idx in range(max_workers):
            pools.add(self._start_thread(idx))

        self.__pool: FrozenSet[threading.Thread] = frozenset(pools)

    def _start_thread(self, idx: int) -> threading.Thread:
        event = threading.Event()
        self.__thread_events.add(event)

        thread_name = "Thread {}".format(idx)

        if self._statistic.name:
            thread_name += " from pool {}".format(self._statistic.name)

        thread = threading.Thread(
            target=self._in_thread,
            name=thread_name.strip(),
            args=(event,),
        )

        thread.daemon = True
        thread.start()
        return thread

    def _in_thread(self, event: threading.Event) -> None:
        self._statistic.threads += 1

        try:
            while True:
                work_item = self.__tasks.get()

                if work_item is None:
                    break

                try:
                    if work_item.loop.is_closed():
                        log.warning(
                            "Event loop is closed. Call %r skipped",
                            work_item.func,
                        )
                        continue

                    work_item(self._statistic)
                finally:
                    del work_item
        finally:
            self._statistic.threads -= 1
            event.set()

    def submit(  # type: ignore
        self, fn: F, *args: Any, **kwargs: Any
    ) -> asyncio.Future:
        """
        Submit blocking function to the pool
        """
        if fn is None or not callable(fn):
            raise ValueError("First argument must be callable")

        loop = asyncio.get_event_loop()
        future: asyncio.Future = loop.create_future()
        self.__futures.add(future)
        future.add_done_callback(self.__futures.remove)

        with self.__write_lock:
            self.__tasks.put_nowait(
                WorkItem(
                    func=fn,
                    args=args,
                    kwargs=kwargs,
                    future=future,
                    loop=loop,
                ),
            )

        self._statistic.submitted += 1
        return future

    # noinspection PyMethodOverriding
    def shutdown(self, wait: bool = True) -> None:  # type: ignore
        for _ in self.__pool:
            self.__tasks.put_nowait(None)

        for f in filter(lambda x: not x.done(), self.__futures):
            f.set_exception(ThreadPoolException("Pool closed"))

        if not wait:
            return

        while not all(e.is_set() for e in self.__thread_events):
            time.sleep(0)

    def _adjust_thread_count(self) -> None:
        raise NotImplementedError

    def __del__(self) -> None:
        self.shutdown()


def run_in_executor(
    func: Callable[..., T],
    executor: ThreadPoolExecutorBase = None,
    args: Any = (),
    kwargs: Any = MappingProxyType({}),
) -> Awaitable[T]:
    try:
        loop = asyncio.get_running_loop()
        return loop.run_in_executor(
            executor, context_partial(func, *args, **kwargs),
        )
    except RuntimeError:
        # In case the event loop is not running right now is
        # returning coroutine to avoid DeprecationWarning in Python 3.10
        async def lazy_wrapper() -> T:
            loop = asyncio.get_running_loop()
            return await loop.run_in_executor(
                executor, context_partial(func, *args, **kwargs),
            )
        return lazy_wrapper()


async def _awaiter(future: asyncio.Future) -> T:
    try:
        result = await future
        return result
    except asyncio.CancelledError as e:
        if not future.done():
            future.set_exception(e)
        raise


def threaded(
    func: Callable[..., T],
) -> Callable[..., Awaitable[T]]:
    if asyncio.iscoroutinefunction(func):
        raise TypeError("Can not wrap coroutine")

    if inspect.isgeneratorfunction(func):
        return threaded_iterable(func)

    @wraps(func)
    def wrap(
        *args: Any, **kwargs: Any
    ) -> Awaitable[T]:
        return run_in_executor(func=func, args=args, kwargs=kwargs)

    return wrap


def run_in_new_thread(
    func: F,
    args: Any = (),
    kwargs: Any = MappingProxyType({}),
    detach: bool = True,
    no_return: bool = False,
    statistic_name: Optional[str] = None,
) -> asyncio.Future:
    loop = asyncio.get_event_loop()
    future = loop.create_future()

    statistic = ThreadPoolStatistic(statistic_name)

    def set_result(result: Any) -> None:
        if future.done() or loop.is_closed():
            return

        future.set_result(result)

    def set_exception(exc: Exception) -> None:
        if future.done() or loop.is_closed():
            return

        future.set_exception(exc)

    @wraps(func)
    def in_thread(target: F) -> None:
        statistic.threads += 1
        statistic.submitted += 1

        try:
            loop.call_soon_threadsafe(
                set_result, target(),
            )
            statistic.success += 1
        except Exception as exc:
            statistic.error += 1
            if loop.is_closed() and no_return:
                return

            elif loop.is_closed():
                log.exception("Uncaught exception from separate thread")
                return

            loop.call_soon_threadsafe(set_exception, exc)
        finally:
            statistic.done += 1
            statistic.threads -= 1

    thread = threading.Thread(
        target=in_thread, name=func.__name__,
        args=(
            context_partial(func, *args, **kwargs),
        ),
        daemon=detach,
    )
    thread.start()
    return future


def threaded_separate(
    func: F,
    detach: bool = True,
) -> Callable[..., Awaitable[Any]]:
    if isinstance(func, bool):
        return partial(threaded_separate, detach=detach)

    if asyncio.iscoroutinefunction(func):
        raise TypeError("Can not wrap coroutine")

    @wraps(func)
    def wrap(*args: Any, **kwargs: Any) -> Any:
        future = run_in_new_thread(
            func, args=args, kwargs=kwargs, detach=detach,
        )

        return _awaiter(future)

    return wrap


def threaded_iterable(
    func: F = None,
    max_size: int = 0,
) -> Any:
    if isinstance(func, int):
        return partial(threaded_iterable, max_size=func)
    if func is None:
        return partial(threaded_iterable, max_size=max_size)

    @wraps(func)
    def wrap(*args: Any, **kwargs: Any) -> Any:
        return IteratorWrapper(
            context_partial(func, *args, **kwargs),  # type: ignore
            max_size=max_size,
        )

    return wrap


class IteratorWrapperSeparate(IteratorWrapper):
    def _run(self) -> Any:
        return run_in_new_thread(self._in_thread)


def threaded_iterable_separate(func: F = None, max_size: int = 0) -> Any:
    if isinstance(func, int):
        return partial(threaded_iterable_separate, max_size=func)
    if func is None:
        return partial(threaded_iterable_separate, max_size=max_size)

    @wraps(func)
    def wrap(*args: Any, **kwargs: Any) -> Any:
        return IteratorWrapperSeparate(
            context_partial(func, *args, **kwargs),  # type: ignore
            max_size=max_size,
        )

    return wrap


class CoroutineWaiter:
    def __init__(
        self, coroutine: Coroutine[Any, Any, T],
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        self.__coro: Coroutine[Any, Any, T] = coroutine
        self.__loop = loop or get_current_loop()
        self.__event = threading.Event()
        self.__result: Optional[T] = None
        self.__exception: Optional[BaseException] = None

    def _on_result(self, task: asyncio.Future) -> None:
        self.__exception = task.exception()
        if self.__exception is None:
            self.__result = task.result()
        self.__event.set()

    def _awaiter(self) -> None:
        task: asyncio.Future = self.__loop.create_task(self.__coro)
        task.add_done_callback(self._on_result)

    def start(self) -> None:
        self.__loop.call_soon_threadsafe(self._awaiter)

    def wait(self) -> Any:
        self.__event.wait()
        if self.__exception is not None:
            raise self.__exception
        return self.__result


def wait_coroutine(
    coro: Coroutine[Any, Any, T],
    loop: Optional[asyncio.AbstractEventLoop] = None,
) -> T:
    waiter = CoroutineWaiter(coro, loop)
    waiter.start()
    return waiter.wait()


def sync_wait_coroutine(
    loop: Optional[asyncio.AbstractEventLoop],
    coro_func: Callable[..., Coroutine[Any, Any, T]],
    *args: Any,
    **kwargs: Any
) -> T:
    return wait_coroutine(coro_func(*args, **kwargs), loop=loop)


def sync_await(
    func: Callable[..., Awaitable[T]],
    *args: Any,
    **kwargs: Any,
) -> T:
    async def awaiter() -> T:
        return await func(*args, **kwargs)
    return wait_coroutine(awaiter())



import logging
import sys
from typing import IO, Any

from colorlog import ColoredFormatter

from ..enum import DateFormat


def color_formatter(
    stream: IO[str] = None,
    date_format: str = None, **_: Any
) -> logging.Handler:

    date_format = (
        date_format if date_format is not None else DateFormat.color.value
    )

    stream = stream or sys.stderr
    handler = logging.StreamHandler(stream)

    fmt = (
        "%(blue)s[T:%(threadName)s]%(reset)s "
        "%(log_color)s%(levelname)s:%(name)s%(reset)s: "
        "%(message_log_color)s%(message)s"
    )

    if date_format:
        fmt = "%(bold_white)s%(bg_black)s%(asctime)s%(reset)s {0}".format(fmt)

    handler.setFormatter(
        ColoredFormatter(
            fmt,
            log_colors={
                "DEBUG": "cyan",
                "INFO": "green",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "red,bg_white",
            },
            secondary_log_colors={
                "message": {
                    "WARNING": "bold",
                    "ERROR": "bold",
                    "CRITICAL": "bold",
                },
            },
            datefmt=date_format,
            reset=True,
            style="%",
        ),
    )

    return handler



import json
import logging
import sys
import traceback
from types import MappingProxyType
from typing import IO, Any, Callable, Dict, List, Union


try:
    from typing import Literal

    StyleType = Union[Literal["%"], Literal["{"], Literal["$"]]
except ImportError:
    StyleType = str     # type: ignore


JSONObjType = Dict[str, Any]
DumpsType = Callable[[JSONObjType, Any], str]


def _dump_json(obj: JSONObjType, *args: Any, **kwargs: Any) -> str:
    kwargs["default"] = repr
    kwargs["ensure_ascii"] = False
    return json.dumps(obj, *args, **kwargs)


class JSONLogFormatter(logging.Formatter):

    LEVELS = MappingProxyType({
        logging.CRITICAL: "crit",
        logging.FATAL: "fatal",
        logging.ERROR: "error",
        logging.WARNING: "warn",
        logging.INFO: "info",
        logging.DEBUG: "debug",
        logging.NOTSET: None,
    })

    FIELD_MAPPING = MappingProxyType({
        "filename": ("code_file", str),
        "funcName": ("code_func", str),
        "lineno": ("code_line", int),
        "module": ("code_module", str),
        "name": ("identifier", str),
        "msg": ("message_raw", str),
        "process": ("pid", int),
        "processName": ("process_name", str),
        "threadName": ("thread_name", str),
    })

    def __init__(
        self, fmt: str = None, datefmt: str = None,
        style: StyleType = "%",
        dumps: DumpsType = _dump_json,
    ):
        super().__init__(
            datefmt=datefmt if datefmt is not Ellipsis else None,
            fmt=fmt, style=style,
        )
        self.dumps = dumps

    def format(self, record: logging.LogRecord) -> str:
        record_dict = MappingProxyType(record.__dict__)

        data: Dict[str, Any] = dict(
            errno=0 if not record.exc_info else 255,
        )

        for key, value in self.FIELD_MAPPING.items():
            mapping, field_type = value

            v = record_dict.get(key)

            if not isinstance(v, field_type):
                v = field_type(v)

            data[mapping] = v

        for key in record_dict:
            if key in data:
                continue
            elif key in self.FIELD_MAPPING:
                continue
            elif key[0] == "_":
                continue

            record_value: Any = record_dict[key]

            if record_value is None:
                continue

            data[key] = record_value

        args: List[Any] = data.pop("args", [])

        for idx, item in enumerate(args):
            data["argument_%d" % idx] = str(item)

        payload = {
            "@fields": data,
            "msg": self.formatMessage(record),
            "level": self.LEVELS[record.levelno],
        }   # type: JSONObjType

        if self.datefmt:
            payload["@timestamp"] = self.formatTime(record, self.datefmt)

        if record.exc_info:
            payload["stackTrace"] = self.formatException(record.exc_info)

        return self.dumps(payload)      # type: ignore

    def formatMessage(self, record: logging.LogRecord) -> str:
        return record.getMessage()

    def formatException(self, exc_info: Any) -> str:
        return "\n".join(traceback.format_exception(*exc_info))

    def formatTime(     # type: ignore
        self, record: logging.LogRecord, datefmt: str = None,
    ) -> Union[int, str]:
        if datefmt == "%s":
            return record.created   # type: ignore
        return super().formatTime(record, datefmt=datefmt)


def json_handler(
    stream: IO[str] = None,
    date_format: str = None,
    **kwargs: Any
) -> logging.Handler:
    log_stream: IO[str] = stream or sys.stdout
    formatter = JSONLogFormatter(datefmt=date_format, **kwargs)
    handler = logging.StreamHandler(log_stream)
    handler.setFormatter(formatter)
    return handler



import asyncio
import inspect
import logging
from asyncio import CancelledError, Event, Future, Lock, wait_for
from inspect import Parameter
from typing import (
    Any, Awaitable, Callable, Iterable, List, NamedTuple, Optional, Union,
)

from .compat import EventLoopMixin
from .counters import Statistic


log = logging.getLogger(__name__)


class Arg(NamedTuple):
    value: Any
    future: Future


class ResultNotSetError(Exception):
    pass


AggFuncHighLevel = Callable[[Any], Awaitable[Iterable]]
AggFuncAsync = Callable[[Arg], Awaitable]
AggFunc = Union[AggFuncHighLevel, AggFuncAsync]


class AggregateStatistic(Statistic):
    leeway_ms: float
    max_count: int
    success: int
    error: int
    done: int


class Aggregator(EventLoopMixin):

    _func: AggFunc
    _max_count: Optional[int]
    _leeway: float
    _first_call_at: Optional[float]
    _args: list
    _futures: List[Future]
    _event: Event
    _lock: Lock

    def __init__(
        self, func: AggFunc, *, leeway_ms: float,
        max_count: Optional[int] = None,
        statistic_name: Optional[str] = None
    ):
        has_variadic_positional = any((
            parameter.kind == Parameter.VAR_POSITIONAL
            for parameter in inspect.signature(func).parameters.values()
        ))
        if not has_variadic_positional:
            raise ValueError(
                "Function must accept variadic positional arguments",
            )

        if max_count is not None and max_count <= 0:
            raise ValueError("max_count must be positive int or None")

        if leeway_ms <= 0:
            raise ValueError("leeway_ms must be positive float")

        self._func = func
        self._max_count = max_count
        self._leeway = leeway_ms / 1000
        self._clear()
        self._statistic = AggregateStatistic(statistic_name)
        self._statistic.leeway_ms = self.leeway_ms
        self._statistic.max_count = max_count or 0

    def _clear(self) -> None:
        self._first_call_at = None
        self._args = []
        self._futures = []
        self._event = Event()
        self._lock = Lock()

    @property
    def max_count(self) -> Optional[int]:
        return self._max_count

    @property
    def leeway_ms(self) -> float:
        return self._leeway * 1000

    @property
    def count(self) -> int:
        return len(self._args)

    async def _execute(self, *, args: list, futures: List[Future]) -> None:
        try:
            results = await self._func(*args)
            self._statistic.success += 1
        except CancelledError:
            # Other waiting tasks can try to finish the job instead.
            raise
        except Exception as e:
            self._set_exception(e, futures)
            self._statistic.error += 1
            return
        finally:
            self._statistic.done += 1

        self._set_results(results, futures)

    def _set_results(self, results: Iterable, futures: List[Future]) -> None:
        for future, result in zip(futures, results):
            if not future.done():
                future.set_result(result)

    def _set_exception(
            self, exc: Exception, futures: List[Future],
    ) -> None:
        for future in futures:
            if not future.done():
                future.set_exception(exc)

    async def aggregate(self, arg: Any) -> Any:
        if self._first_call_at is None:
            self._first_call_at = self.loop.time()
        first_call_at = self._first_call_at

        args: list = self._args
        futures: List[Future] = self._futures
        event: Event = self._event
        lock: Lock = self._lock
        args.append(arg)
        future: Future = Future()
        futures.append(future)

        if self.count == self.max_count:
            event.set()
            self._clear()
        else:
            # Waiting for max_count requests or a timeout
            try:
                await wait_for(
                    event.wait(),
                    timeout=first_call_at + self._leeway - self.loop.time(),
                )
            except asyncio.TimeoutError:
                log.debug(
                    "Aggregation timeout of %s for batch started at %.4f "
                    "with %d calls after %.2f ms",
                    self._func.__name__, first_call_at, len(futures),
                    (self.loop.time() - first_call_at) * 1000,
                )

        # Clear only if not cleared already
        if args is self._args:
            self._clear()

        # Trying to acquire the lock to execute the aggregated function
        async with lock:
            if not future.done():
                await self._execute(args=args, futures=futures)
        await future
        return future.result()


class AggregatorAsync(Aggregator):

    async def _execute(self, *, args: list, futures: List[Future]) -> None:
        args = [
            Arg(value=arg, future=future)
            for arg, future in zip(args, futures)
        ]
        try:
            await self._func(*args)
            self._statistic.success += 1
        except CancelledError:
            # Other waiting tasks can try to finish the job instead.
            raise
        except Exception as e:
            self._set_exception(e, futures)
            self._statistic.error += 1
            return
        finally:
            self._statistic.done += 1

        # Validate that all results/exceptions are set by the func
        for future in futures:
            if not future.done():
                future.set_exception(ResultNotSetError)


def aggregate(leeway_ms: float, max_count: Optional[int] = None) -> Callable:
    """
    Parametric decorator that aggregates multiple
    (but no more than ``max_count`` defaulting to ``None``) single-argument
    executions (``res1 = await func(arg1)``, ``res2 = await func(arg2)``, ...)
    of an asynchronous function with variadic positional arguments
    (``async def func(*args, pho=1, bo=2) -> Iterable``) into its single
    execution with multiple positional arguments
    (``res1, res2, ... = await func(arg1, arg2, ...)``) collected within a time
    window ``leeway_ms``.

    .. note::

        ``func`` must return a sequence of values of length equal to the
        number of arguments (and in the same order).

    .. note::

        if some unexpected error occurs, exception is propagated to each
        future; to set an individual error for each aggregated call refer
        to ``aggregate_async``.

    :param leeway_ms: The maximum approximate delay between the first
           collected argument and the aggregated execution.
    :param max_count: The maximum number of arguments to call decorated
           function with. Default ``None``.

    :return:
    """
    def decorator(func: AggFuncHighLevel) -> Callable[[Any], Awaitable]:
        aggregator = Aggregator(
            func, max_count=max_count, leeway_ms=leeway_ms,
        )
        return aggregator.aggregate
    return decorator


def aggregate_async(
        leeway_ms: float, max_count: Optional[int] = None,
) -> Callable:
    """
    Same as ``aggregate``, but with ``func`` arguments of type ``Arg``
    containing ``value`` and ``future`` attributes instead. In this setting
    ``func`` is responsible for setting individual results/exceptions for all
    of the futures or throwing an exception (it will propagate to futures
    automatically). If ``func`` mistakenly does not set a result of some
    future, then, ``ResultNotSetError`` exception is set.

    :return:
    """
    def decorator(func: AggFuncAsync) -> Callable[[Any], Awaitable]:
        aggregator = AggregatorAsync(
            func, max_count=max_count, leeway_ms=leeway_ms,
        )
        return aggregator.aggregate
    return decorator



import logging
import os
import signal
from abc import ABC, abstractmethod
from multiprocessing import Event, Process, synchronize
from threading import Lock
from typing import Any, Callable, Dict, Optional

from aiomisc.periodic import PeriodicCallback
from aiomisc.service.base import Service
from aiomisc_log import LOG_FORMAT, LOG_LEVEL, LogFormat, basic_config


log = logging.getLogger(__name__)


def _process_inner(
    function: Callable[..., Any],
    log_level: str,
    log_format: str,
    start_event: synchronize.Event,
    stop_event: synchronize.Event,
    **kwargs: Any
) -> None:
    basic_config(level=log_level, log_format=log_format)
    start_event.set()
    try:
        function(**kwargs)
    finally:
        stop_event.set()


class ProcessService(Service):
    name: Optional[str] = None
    process: Process
    process_stop_timeout: int = 3

    _process_start_event: synchronize.Event
    _process_stop_event: synchronize.Event
    _lock: Lock
    _instance_params: Dict[str, Any]

    def get_process_kwargs(self) -> Dict[str, Any]:
        return {}

    @abstractmethod
    def in_process(self) -> Any:
        pass

    def _is_alive(self) -> bool:
        with self._lock:
            if not hasattr(self, "process"):
                return False
            if not self.process.pid:
                return False
            return self.process.is_alive()

    async def start(self) -> Any:
        self._lock = Lock()
        self._process_start_event = Event()
        self._process_stop_event = Event()

        with self._lock:
            log_level = (
                log.getEffectiveLevel()
                if LOG_LEVEL is None
                else LOG_LEVEL.get()
            )
            log_format = (
                LogFormat.default()
                if LOG_FORMAT is None
                else LOG_FORMAT.get().value
            )

            process = Process(
                target=_process_inner,
                args=(
                    self.in_process,
                    log_level,
                    log_format,
                    self._process_start_event,
                    self._process_stop_event,
                ),
                kwargs=self.get_process_kwargs(),
                name=self.name,
            )

            process.start()

            await self.loop.run_in_executor(
                None, self._process_start_event.wait,
            )
            self.process = process

    def __repr__(self) -> str:
        pid: Optional[int] = None
        if hasattr(self, "process"):
            pid = self.process.pid

        return "<{} object at {}: name={!r}, pid={}>".format(
            self.__class__.__name__, hex(id(self)), self.name, pid,
        )

    async def stop(self, exception: Exception = None) -> Any:
        if not self._is_alive():
            return

        process = self.process
        del self.process

        if not process.pid:
            return

        os.kill(process.pid, signal.SIGINT)
        stop_result: bool = await self.loop.run_in_executor(
            None, self._process_stop_event.wait, self.process_stop_timeout,
        )

        if stop_result:
            stop_result = (
                await self.loop.run_in_executor(
                    None, process.join, self.process_stop_timeout,
                )
            ) is not None

        if not stop_result and process.is_alive():
            process.kill()
            log.warning(
                f"The process {process.pid} didn't stop for "
                f"{self.process_stop_timeout} seconds "
                f"and had been killed.",
            )


class RespawningProcessService(ProcessService, ABC):
    process_poll_timeout: int = 5

    _supervisor: PeriodicCallback

    async def __supervise(self) -> None:
        if not hasattr(self, "process"):
            return

        if await self.loop.run_in_executor(None, self.process.is_alive):
            return

        log.info(
            "Process in service %r exited with code %r, respawning.",
            self, self.process.exitcode,
        )
        await super().start()

    async def start(self) -> None:
        await super().start()
        self._supervisor = PeriodicCallback(self.__supervise)
        self._supervisor.start(
            self.process_poll_timeout,
        )

    async def stop(self, exception: Exception = None) -> Any:
        await self._supervisor.stop()
        await super().stop(exception)


__all__ = ("ProcessService", "RespawningProcessService")



import sys

from aiomisc_log import basic_config

from . import AddressType
from .protocol import FileIOProtocol
from .worker import bad_initializer, worker


def worker_inner() -> None:
    proto = FileIOProtocol(sys.stdin.buffer)
    log_level, log_format = proto.receive()
    basic_config(level=log_level, log_format=log_format)

    address: AddressType = proto.receive()
    cookie: bytes = proto.receive()
    worker_id: bytes = proto.receive()
    initializer, initializer_args, initializer_kwargs = proto.receive()
    sys.stdin.close()
    del proto

    if initializer is not None:
        try:
            initializer(*initializer_args, **initializer_kwargs)
        except BaseException as e:
            bad_initializer(address, cookie, worker_id, e)
            raise SystemExit(0)

    del initializer
    del initializer_args
    del initializer_kwargs

    return worker(address, cookie, worker_id)


if __name__ == "__main__":
    worker_inner()



import asyncio
from functools import wraps
from typing import Any, Awaitable, Callable, TypeVar, Union


T = TypeVar("T")
Number = Union[int, float]
FuncType = Callable[..., Awaitable[T]]


def timeout(value: Number) -> Callable[[FuncType], FuncType]:
    def decorator(
        func: FuncType,
    ) -> FuncType:
        if not asyncio.iscoroutinefunction(func):
            raise TypeError("Function is not a coroutine function")

        @wraps(func)
        async def wrap(*args: Any, **kwargs: Any) -> T:
            return await asyncio.wait_for(
                func(*args, **kwargs),
                timeout=value,
            )
        return wrap
    return decorator



import logging
import socket
import struct
import traceback
import uuid
from enum import IntEnum, unique
from io import BytesIO
from pathlib import Path
from types import MappingProxyType
from typing import Any, BinaryIO


@unique
class Facility(IntEnum):
    KERN = 0
    USER = 1
    MAIL = 2
    DAEMON = 3
    AUTH = 4
    SYSLOG = 5
    LPR = 6
    NEWS = 7
    UUCP = 8
    CLOCK_DAEMON = 9
    AUTHPRIV = 10
    FTP = 11
    NTP = 12
    AUDIT = 13
    ALERT = 14
    CRON = 15
    LOCAL0 = 16
    LOCAL1 = 17
    LOCAL2 = 18
    LOCAL3 = 19
    LOCAL4 = 20
    LOCAL5 = 21
    LOCAL6 = 22
    LOCAL7 = 23


class JournaldLogHandler(logging.Handler):
    LEVELS = MappingProxyType({
        logging.CRITICAL: 2,
        logging.DEBUG: 7,
        logging.FATAL: 0,
        logging.ERROR: 3,
        logging.INFO: 6,
        logging.NOTSET: 16,
        logging.WARNING: 4,
    })

    VALUE_LEN_STRUCT = struct.Struct("<Q")
    SOCKET_PATH = Path("/run/systemd/journal/socket")

    RECORD_FIELDS_MAP = MappingProxyType({
        "args": "arguments",
        "created": None,
        "exc_info": None,
        "exc_text": None,
        "filename": None,
        "funcName": None,
        "levelname": None,
        "levelno": None,
        "lineno": None,
        "message": None,
        "module": None,
        "msecs": None,
        "msg": "message_raw",
        "name": "logger_name",
        "pathname": None,
        "process": "pid",
        "processName": "process_name",
        "relativeCreated": None,
        "thread": "thread_id",
        "threadName": "thread_name",
    })

    __slots__ = ("__facility", "socket", "__identifier")

    @staticmethod
    def _encode_short(key: str, value: Any) -> bytes:
        return "{}={}\n".format(key.upper(), value).encode()

    @classmethod
    def _encode_long(cls, key: str, value: bytes) -> bytes:
        length = cls.VALUE_LEN_STRUCT.pack(len(value))
        return key.upper().encode() + b"\n" + length + value + b"\n"

    @classmethod
    def pack(cls, fp: BinaryIO, key: str, value: Any) -> None:
        if not value:
            return
        elif isinstance(value, (int, float)):
            fp.write(cls._encode_short(key, value))
            return
        elif isinstance(value, str):
            if "\n" in value:
                fp.write(cls._encode_long(key, value.encode()))
                return
            fp.write(cls._encode_short(key, value))
            return
        elif isinstance(value, bytes):
            fp.write(cls._encode_long(key, value))
            return
        elif isinstance(value, (list, tuple)):
            for idx, item in enumerate(value):
                cls.pack(fp, "{}_{}".format(key, idx), item)
            return
        elif isinstance(value, dict):
            for d_key, d_value in value.items():
                cls.pack(fp, "{}_{}".format(key, d_key), d_value)
            return

        cls.pack(fp, key, str(value).encode())
        return

    def __init__(
        self, identifier: str = None,
        facility: int = Facility.LOCAL7,
    ):
        super().__init__()
        self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
        self.socket.connect(str(self.SOCKET_PATH))
        self.__identifier = identifier
        self.__facility = int(facility)

    @staticmethod
    def _to_usec(ts: float) -> int:
        return int(ts * 1000000)

    def emit(self, record: logging.LogRecord) -> None:
        with BytesIO() as fp:
            message_id = uuid.uuid1().hex

            self.pack(fp, "message", self.format(record))
            self.pack(fp, "message_id", message_id)

            if record.exc_info:
                exc_type, exc_value, exc_tb = record.exc_info
                self.pack(
                    fp, "exception", {
                        "type": exc_type,
                        "value": exc_value,
                    },
                )
                tb_message = "\n".join(
                    traceback.format_exception(*record.exc_info),
                )
                self.pack(fp, "traceback", tb_message)

            self.pack(fp, "priority", self.LEVELS[record.levelno])
            self.pack(fp, "syslog_facility", self.__facility)
            self.pack(fp, "syslog_identifier", self.__identifier)
            self.pack(
                fp, "code", "{}.{}:{}".format(
                    record.module, record.funcName, record.lineno,
                ),
            )

            self.pack(
                fp, "code", {
                    "func": record.funcName,
                    "file": record.pathname,
                    "line": record.lineno,
                    "module": record.module,
                },
            )

            self.pack(fp, "created_usec", self._to_usec(record.created))
            self.pack(
                fp, "relative_usec", self._to_usec(record.relativeCreated),
            )

            source = dict(record.__dict__)

            for field, name in self.RECORD_FIELDS_MAP.items():
                value = source.pop(field, None)
                if name is None or value is None:
                    continue
                self.pack(fp, name, value)

            self.pack(fp, "extra", source)

            self.socket.sendall(fp.getvalue())


def journald_formatter(**_: Any) -> logging.Handler:
    handler = JournaldLogHandler()
    formatter = logging.Formatter("%(message)s")
    handler.setFormatter(formatter)
    return handler



import logging
import os
from types import MappingProxyType
from typing import Callable, Mapping


def setup_plugins() -> Mapping[str, Callable]:
    if os.getenv("AIOMISC_NO_PLUGINS"):
        return MappingProxyType({})

    import pkg_resources

    plugins = {}

    for entry_point in pkg_resources.iter_entry_points("aiomisc.plugins"):
        plugins[entry_point.name] = entry_point.load()

    logger = logging.getLogger(__name__)
    for name, plugin in plugins.items():
        try:
            logger.debug("Trying to load %r %r", name, plugin)
            plugin.setup()
        except:  # noqa
            logger.exception("Error on %s aiomisc plugin setup", name)
            raise

    return MappingProxyType(plugins)


plugins: Mapping[str, Callable] = setup_plugins()


__all__ = ("plugins",)

if __name__ == "__main__":
    from aiomisc_log import LogFormat, basic_config

    basic_config(log_format=LogFormat.plain)
    logging.info("Available %s plugins.", len(plugins))

    for name in plugins:
        print(name)



import asyncio
from concurrent.futures import Executor
from functools import partial, total_ordering
from pathlib import Path
from typing import Any, Awaitable, Callable, Generator, Optional, TypeVar, Union

from .compat import EventLoopMixin
from .thread_pool import threaded


T = TypeVar("T")


def proxy_method_async(
    name: str,
    in_executor: bool = True,
) -> Callable[..., Any]:
    def wrap_to_future(
        loop: asyncio.AbstractEventLoop,
        func: Callable[..., T],
        *args: Any, **kwargs: Any
    ) -> asyncio.Future:
        future = loop.create_future()

        def _inner():   # type: ignore
            try:
                return future.set_result(func(*args, **kwargs))
            except Exception as e:
                return future.set_exception(e)

        loop.call_soon(_inner)
        return future

    def wrap_to_thread(
        loop: asyncio.AbstractEventLoop, func: Callable[..., T],
        executor: Executor,
        *args: Any, **kwargs: Any
    ) -> Awaitable[T]:
        callee = partial(func, *args, **kwargs)
        # noinspection PyTypeChecker
        return loop.run_in_executor(executor, callee)

    async def method(self, *args, **kwargs):    # type: ignore
        func = getattr(self.fp, name)

        if in_executor:
            return await wrap_to_thread(
                self.loop, func, self.executor, *args, **kwargs
            )

        return await wrap_to_future(self.loop, func, *args, **kwargs)

    method.__name__ = name
    return method


def proxy_method(name: str) -> Callable[..., Awaitable[Any]]:
    def method(self: Any, *args: Any, **kwargs: Any) -> Any:
        return getattr(self.fp, name)(*args, **kwargs)

    method.__name__ = name
    return method


def proxy_property(name: str) -> property:
    def fset(self: Any, value: Any) -> None:
        setattr(self.fp, name, value)

    def fget(self: Any) -> Any:
        return getattr(self.fp, name)

    def fdel(self: Any) -> None:
        delattr(self.fp, name)

    return property(fget, fset, fdel)


@total_ordering
class AsyncFileIOBase(EventLoopMixin):
    __slots__ = (
        "__opener", "fp", "executor", "__iterator_lock",
    ) + EventLoopMixin.__slots__

    opener = staticmethod(threaded(open))

    def __init__(
        self, fname: Union[str, Path], mode: str = "r",
        executor: Executor = None, *args: Any,
        loop: Optional[asyncio.AbstractEventLoop] = None, **kwargs: Any
    ):
        self._loop = loop
        self.fp = None
        self.executor = executor
        self.__opener = partial(self.opener, fname, mode, *args, **kwargs)
        self.__iterator_lock = asyncio.Lock()

    def closed(self) -> bool:
        if self.fp is None:
            raise RuntimeError("file is not opened")
        return self.fp.closed

    async def open(self) -> None:
        if self.fp is not None:
            return

        self.fp = await self.__opener()     # type: ignore

    def __await__(self) -> Generator[Any, Any, "AsyncFileIOBase"]:
        yield from self.open().__await__()
        return self

    async def __aenter__(self) -> "AsyncFileIOBase":
        await self.open()
        return self

    async def __aexit__(
        self, exc_type: Any,
        exc_val: Any, exc_tb: Any,
    ) -> None:
        if self.fp is None:
            raise RuntimeError("file is not opened")

        await self.loop.run_in_executor(
            None, self.fp.__exit__, exc_type, exc_val, exc_tb,
        )

    def __del__(self) -> None:
        if not self.fp or self.fp.closed:
            return

        self.fp.close()
        del self.fp

    def __aiter__(self) -> "AsyncFileIOBase":
        return self

    async def __anext__(self) -> Union[str, bytes]:
        async with self.__iterator_lock:
            line = await self.readline()

        if not len(line):
            raise StopAsyncIteration

        return line

    def __eq__(self, other: Any) -> bool:
        return (
            self.__class__, self.fp.__eq__(other),
        ) == (
            other.__class__, self.fp.__eq__(other),
        )

    def __lt__(self, other: Any) -> bool:
        return self.fp < other.fp

    def __hash__(self) -> int:
        return hash((self.__class__, self.fp))

    fileno = proxy_method("fileno")
    isatty = proxy_method("isatty")

    mode = proxy_property("mode")
    name = proxy_property("name")

    close = proxy_method_async("close")
    detach = proxy_method_async("detach")
    flush = proxy_method_async("flush")
    peek = proxy_method_async("peek")
    raw = proxy_property("raw")
    read = proxy_method_async("read")
    read1 = proxy_method_async("read1")
    readinto = proxy_method_async("readinto")
    readinto1 = proxy_method_async("readinto1")
    readline = proxy_method_async("readline")
    readlines = proxy_method_async("readlines")
    seek = proxy_method_async("seek")
    peek = proxy_method_async("peek")
    truncate = proxy_method_async("truncate")
    write = proxy_method_async("write")
    writelines = proxy_method_async("writelines")

    tell = proxy_method_async("tell", in_executor=False)
    readable = proxy_method_async("readable", in_executor=False)
    seekable = proxy_method_async("seekable", in_executor=False)
    writable = proxy_method_async("writable", in_executor=False)


class AsyncTextFileIOBase:
    newlines = proxy_property("newlines")
    errors = proxy_property("errors")
    line_buffering = proxy_property("line_buffering")
    encoding = proxy_property("encoding")
    buffer = proxy_property("buffer")


class AsyncBytesFileIOBase:
    pass


class AsyncTextFileIO(AsyncFileIOBase, AsyncTextFileIOBase):
    pass


class AsyncBytesFileIO(AsyncFileIOBase, AsyncBytesFileIOBase):
    pass


AsyncFileT = Union[
    AsyncBytesFileIO,
    AsyncTextFileIO,
]


def async_open(
    fname: Union[str, Path], mode: str = "r",
    *args: Any, **kwargs: Any
) -> AsyncFileT:
    if "b" in mode:
        return AsyncBytesFileIO(
            fname, mode, *args, **kwargs
        )
    return AsyncTextFileIO(fname, mode, *args, **kwargs)



import asyncio
import inspect
from abc import abstractmethod
from collections import deque
from concurrent.futures import Executor
from queue import Empty as QueueEmpty
from queue import Queue
from time import time
from types import TracebackType
from typing import (
    Any, AsyncIterator, Awaitable, Callable, Deque, Generator, NoReturn,
    Optional, Type, TypeVar, Union,
)
from weakref import finalize

from aiomisc.compat import EventLoopMixin
from aiomisc.counters import Statistic


T = TypeVar("T")
R = TypeVar("R")

GenType = Generator[T, R, None]
FuncType = Callable[[], GenType]


class ChannelClosed(RuntimeError):
    pass


class QueueWrapperBase:
    @abstractmethod
    def put(self, item: Any) -> None:
        raise NotImplementedError

    def get(self) -> Any:
        raise NotImplementedError


class DequeWrapper(QueueWrapperBase):
    __slots__ = "queue",

    def __init__(self) -> None:
        self.queue: Deque[Any] = deque()

    def get(self) -> Any:
        if not self.queue:
            raise QueueEmpty
        return self.queue.popleft()

    def put(self, item: Any) -> None:
        return self.queue.append(item)


class QueueWrapper(QueueWrapperBase):
    __slots__ = "queue",

    def __init__(self, max_size: int) -> None:
        self.queue: Queue = Queue(maxsize=max_size)

    def put(self, item: Any) -> None:
        return self.queue.put(item)

    def get(self) -> Any:
        return self.queue.get_nowait()


def make_queue(max_size: int = 0) -> QueueWrapperBase:
    if max_size > 0:
        return QueueWrapper(max_size)
    return DequeWrapper()


class FromThreadChannel:
    SLEEP_LOW_THRESHOLD = 0.0001
    SLEEP_DIFFERENCE_DIVIDER = 10

    __slots__ = ("queue", "__closed", "__last_received_item")

    def __init__(self, maxsize: int = 0):
        self.queue: QueueWrapperBase = make_queue(max_size=maxsize)
        self.__closed = False
        self.__last_received_item: float = time()

    def close(self) -> None:
        self.__closed = True

    @property
    def is_closed(self) -> bool:
        return self.__closed

    def __enter__(self) -> "FromThreadChannel":
        return self

    def __exit__(
        self, exc_type: Type[Exception],
        exc_val: Exception, exc_tb: TracebackType,
    ) -> None:
        self.close()

    def put(self, item: Any) -> None:
        if self.is_closed:
            raise ChannelClosed

        self.queue.put(item)
        self.__last_received_item = time()

    def _compute_sleep_time(self) -> Union[float, int]:
        if self.__last_received_item < 0:
            return 0

        delta = time() - self.__last_received_item

        if delta > 1:
            return 1

        sleep_time = delta / self.SLEEP_DIFFERENCE_DIVIDER

        if sleep_time < self.SLEEP_LOW_THRESHOLD:
            return 0
        return sleep_time

    def __await__(self) -> Any:
        while True:
            try:
                res = self.queue.get()
                return res
            except QueueEmpty:
                if self.is_closed:
                    raise ChannelClosed

                sleep_time = self._compute_sleep_time()
                yield from asyncio.sleep(sleep_time).__await__()

    async def get(self) -> Any:
        return await self


class IteratorWrapperStatistic(Statistic):
    started: int
    queue_size: int
    queue_length: int
    yielded: int
    enqueued: int


class IteratorWrapper(AsyncIterator, EventLoopMixin):
    __slots__ = (
        "__channel",
        "__close_event",
        "__gen_func",
        "__gen_task",
        "_statistic",
        "executor",
    ) + EventLoopMixin.__slots__

    def __init__(
        self, gen_func: FuncType,
        loop: asyncio.AbstractEventLoop = None,
        max_size: int = 0, executor: Executor = None,
        statistic_name: Optional[str] = None,
    ):

        self._loop = loop
        self.executor = executor

        self.__close_event = asyncio.Event()
        self.__channel: FromThreadChannel = FromThreadChannel(maxsize=max_size)
        self.__gen_task: Optional[asyncio.Future] = None
        self.__gen_func: Callable = gen_func
        self._statistic = IteratorWrapperStatistic(statistic_name)
        self._statistic.queue_size = max_size

    @property
    def closed(self) -> bool:
        return self.__channel.is_closed

    @staticmethod
    def __throw(_: Any) -> NoReturn:
        pass

    def _in_thread(self) -> None:
        self._statistic.started += 1
        with self.__channel:
            try:
                gen = iter(self.__gen_func())

                throw = self.__throw
                if inspect.isgenerator(gen):
                    throw = gen.throw   # type: ignore

                while not self.closed:
                    item = next(gen)
                    try:
                        self.__channel.put((item, False))
                    except Exception as e:
                        throw(e)
                        self.__channel.close()
                        break
                    finally:
                        del item

                    self._statistic.enqueued += 1
            except StopIteration:
                return
            except Exception as e:
                if self.closed:
                    return
                self.__channel.put((e, True))
            finally:
                self._statistic.started -= 1
                self.loop.call_soon_threadsafe(self.__close_event.set)

    def close(self) -> Awaitable[None]:
        self.__channel.close()
        return asyncio.ensure_future(self.wait_closed())

    async def wait_closed(self) -> None:
        await self.__close_event.wait()
        if self.__gen_task:
            await asyncio.gather(self.__gen_task, return_exceptions=True)

    def _run(self) -> Any:
        return self.loop.run_in_executor(
            self.executor, self._in_thread,
        )

    def __aiter__(self) -> AsyncIterator[Any]:
        if not self.loop.is_running():
            raise RuntimeError("Event loop is not running")

        if self.__gen_task is None:
            gen_task = self._run()
            if gen_task is None:
                raise RuntimeError("Iterator task was not created")
            self.__gen_task = gen_task
        return IteratorProxy(self, self.close)

    async def __anext__(self) -> Awaitable[T]:
        try:
            item, is_exc = await self.__channel.get()
        except ChannelClosed:
            await self.wait_closed()
            raise StopAsyncIteration

        if is_exc:
            await self.close()
            raise item from item

        self._statistic.yielded += 1
        return item

    async def __aenter__(self) -> "IteratorWrapper":
        return self

    async def __aexit__(
        self, exc_type: Any, exc_val: Any,
        exc_tb: Any,
    ) -> None:
        if self.closed:
            return

        await self.close()


class IteratorProxy(AsyncIterator):
    def __init__(
        self, iterator: AsyncIterator,
        finalizer: Callable[[], Any],
    ):
        self.__iterator = iterator
        finalize(self, finalizer)

    def __anext__(self) -> Awaitable[Any]:
        return self.__iterator.__anext__()



import abc
import asyncio
import logging
import os
import platform
import socket
import sys
import warnings
from asyncio.events import get_event_loop
from contextlib import contextmanager, suppress
from functools import partial, wraps
from inspect import isasyncgenfunction
from typing import Awaitable, Callable, NamedTuple, Optional, Tuple, Type, Union
from unittest.mock import MagicMock

import pytest

import aiomisc
from aiomisc.utils import bind_socket, sock_set_reuseport
from aiomisc_log import LOG_LEVEL, basic_config


log = logging.getLogger("aiomisc_pytest")
asyncio.get_event_loop = MagicMock(asyncio.get_event_loop)
asyncio.get_event_loop.side_effect = get_event_loop


try:
    import uvloop  # type: ignore
except ImportError:
    uvloop = None


def delayed_future(
    timeout: Union[int, float], result: bool = True,
) -> asyncio.Future:

    loop = asyncio.get_event_loop()

    def resolve(f: asyncio.Future) -> None:
        nonlocal result

        if f.done():
            return
        f.set_result(result)

    future = loop.create_future()
    handle = loop.call_later(timeout, resolve, future)
    future.add_done_callback(lambda _: handle.cancel())

    return future


ProxyProcessorType = Callable[[bytes], Awaitable[bytes]]
DelayType = Union[int, float]


class Delay:
    __slots__ = "__timeout", "future", "lock"

    def __init__(self) -> None:
        self.__timeout: Union[int, float] = 0
        self.future = None
        self.lock = asyncio.Lock()

    @property
    def timeout(self) -> Union[int, float]:
        return self.__timeout

    @timeout.setter
    def timeout(self, value: DelayType):
        assert isinstance(value, (int, float))
        assert value >= 0

        self.__timeout = value

        if self.future and not self.future.done():
            self.future.set_result(True)

    async def wait(self):
        if self.__timeout == 0:
            return

        async with self.lock:
            try:
                self.future = delayed_future(self.__timeout)
                await self.future
            finally:
                self.future = None


class TCPProxyClient:
    __slots__ = (
        "client_reader", "client_writer",
        "server_reader", "server_writer",
        "chunk_size", "tasks", "loop",
        "read_delay", "write_delay", "closing",
        "__processors", "__client_repr", "__server_repr",
        "buffered",
    )

    @staticmethod
    async def _blank_processor(body: bytes) -> bytes:
        return body

    def __init__(
        self, client_reader: asyncio.StreamReader,
        client_writer: asyncio.StreamWriter,
        chunk_size: int = 64 * 1024, buffered: bool = False,
    ):

        self.loop = asyncio.get_event_loop()
        self.client_reader = client_reader  # type: asyncio.StreamReader
        self.client_writer = client_writer  # type: asyncio.StreamWriter

        self.server_reader = None  # type: t.Optional[asyncio.StreamReader]
        self.server_writer = None  # type: t.Optional[asyncio.StreamWriter]

        self.tasks = ()  # type: t.Iterable[asyncio.Task]
        self.chunk_size = chunk_size  # type: int
        self.read_delay = Delay()
        self.write_delay = Delay()

        self.closing = self.loop.create_future()  # type: asyncio.Future
        self.__processors = {
            "read": self._blank_processor,
            "write": self._blank_processor,
        }  # type: t.Dict[str, t.Callable[[], ProxyProcessorType]]

        self.buffered = bool(buffered)

        self.__client_repr = None
        self.__server_repr = None

    @property
    def read_processor(self) -> Callable[[], ProxyProcessorType]:
        return self.__processors["read"]

    @read_processor.setter
    def read_processor(self, value: Optional[ProxyProcessorType]) -> None:
        if value is None:
            self.__processors["read"] = self._blank_processor
            return

        self.__processors["read"] = aiomisc.awaitable(value)

    @property
    def write_processor(self) -> Callable[[], ProxyProcessorType]:
        return self.__processors["write"]

    @write_processor.setter
    def write_processor(self, value: Optional[ProxyProcessorType]) -> None:
        if value is None:
            self.__processors["write"] = self._blank_processor
            return
        self.__processors["write"] = aiomisc.awaitable(value)

    def __repr__(self):
        return "<%s[%x]: %s => %s>" % (
            self.__class__.__name__, id(self),
            self.__client_repr, self.__server_repr,
        )

    if sys.version_info < (3, 7):
        @staticmethod
        async def _close_writer(writer: asyncio.StreamWriter):
            writer.close()
    else:
        @staticmethod
        async def _close_writer(writer: asyncio.StreamWriter):
            writer.close()
            await writer.wait_closed()

    async def pipe(
        self, reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        processor: str,
        delay: Delay,
    ) -> None:
        try:
            while not reader.at_eof():
                chunk = await reader.read(self.chunk_size)

                if not chunk:
                    break

                if delay.timeout > 0:
                    log.debug(
                        "%r sleeping %.3f seconds on %s",
                        self, delay.timeout, processor,
                    )

                    await delay.wait()

                writer.write(await self.__processors[processor](chunk))

                if not self.buffered:
                    await writer.drain()
        finally:
            await self._close_writer(writer)

    async def connect(self, target_host: str, target_port: int) -> None:
        log.debug("Establishing connection for %r", self)

        self.server_reader, self.server_writer = await asyncio.open_connection(
            host=target_host, port=target_port,
        )

        self.__client_repr = ":".join(
            map(str, self.client_writer.get_extra_info("peername")[:2]),
        )
        self.__server_repr = ":".join(
            map(str, self.server_writer.get_extra_info("peername")[:2]),
        )

        self.tasks = (
            self.loop.create_task(
                self.pipe(
                    self.server_reader,
                    self.client_writer,
                    "write",
                    self.write_delay,
                ),
            ),
            self.loop.create_task(
                self.pipe(
                    self.client_reader,
                    self.server_writer,
                    "read",
                    self.read_delay,
                ),
            ),
        )

    async def close(self):
        log.debug("Closing %r", self)
        if self.closing.done():
            return

        await aiomisc.cancel_tasks(self.tasks)
        self.loop.call_soon(self.closing.set_result, True)
        await self.closing


class TCPProxy:
    DEFAULT_TIMEOUT = 30

    __slots__ = (
        "proxy_port", "clients", "server", "proxy_host",
        "target_port", "target_host", "listen_host", "read_delay",
        "write_delay", "read_processor", "write_processor", "buffered",
    )

    def __init__(
            self, target_host: str, target_port: int,
            listen_host: str = "127.0.0.1", buffered: bool = True,
    ):
        self.target_port = target_port
        self.target_host = target_host
        self.proxy_host = listen_host
        self.proxy_port = get_unused_port()
        self.read_delay = 0
        self.write_delay = 0
        self.buffered = buffered

        self.clients = set()  # type: t.Set[TCPProxyClient]
        self.server = None  # type: t.Optional[asyncio.AbstractServer]

        self.read_processor = None  # type: t.Optional[ProxyProcessorType]
        self.write_processor = None  # type: t.Optional[ProxyProcessorType]

    def __repr__(self):
        return "<%s[%x]: tcp://%s:%s => tcp://%s:%s>" % (
            self.__class__.__name__, id(self),
            self.proxy_host, self.proxy_port,
            self.target_host, self.target_port,
        )

    async def start(self, timeout=None) -> asyncio.AbstractServer:
        log.debug("Starting %r", self)
        self.server = await asyncio.wait_for(
            asyncio.start_server(
                self._handle_client,
                host=self.proxy_host,
                port=self.proxy_port,
            ), timeout=timeout,
        )
        return self.server

    ClientType = Tuple[asyncio.StreamReader, asyncio.StreamWriter]

    async def create_client(self) -> ClientType:
        log.debug("Creating client for %r", self)
        return await asyncio.open_connection(
            self.proxy_host, self.proxy_port,
        )

    async def __aenter__(self):
        if self.server is None:
            await self.start(timeout=self.DEFAULT_TIMEOUT)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close(timeout=self.DEFAULT_TIMEOUT)

    async def close(self, timeout=None):
        async def close():
            await self.disconnect_all()
            self.server.close()
            await self.server.wait_closed()
        await asyncio.wait_for(close(), timeout=timeout)

    def set_delay(self, read_delay: DelayType, write_delay: DelayType = 0):
        log.debug("Setting delay [R/W %f %f]", read_delay, write_delay)

        for client in self.clients:
            log.debug(
                "Applying delays [R/W: %f %f] for %r",
                read_delay, write_delay, client,
            )
            client.read_delay.timeout = read_delay
            client.write_delay.timeout = write_delay

        self.read_delay = read_delay
        self.write_delay = write_delay

    def set_content_processors(
            self, read: Optional[ProxyProcessorType],
            write: Optional[ProxyProcessorType],
    ):
        log.debug(
            "Setting content processors for %r: read=%r write=%r",
            self, read, write,
        )
        for client in self.clients:
            log.debug(
                "Applying context processors for %r: read=%r write=%r",
                client, read, write,
            )
            client.read_processor = read
            client.write_processor = write

        self.read_processor = read
        self.write_processor = write

    def disconnect_all(self) -> asyncio.Future:
        log.debug(
            "Disconnecting %s clients of %r", len(self.clients), self,
        )
        return asyncio.ensure_future(
            asyncio.gather(
                *[client.close() for client in self.clients],
                return_exceptions=True
            ),
        )

    async def _handle_client(
            self, reader: asyncio.StreamReader,
            writer: asyncio.StreamWriter,
    ):
        client = TCPProxyClient(reader, writer, buffered=self.buffered)
        self.clients.add(client)

        client.read_delay.timeout = self.read_delay
        client.write_delay.timeout = self.write_delay
        client.read_processor = self.read_processor
        client.write_processor = self.write_processor

        await client.connect(self.target_host, self.target_port)

        client.closing.add_done_callback(
            lambda _: self.clients.remove(client),
        )

    @contextmanager
    def slowdown(self, read_delay: DelayType = 0, write_delay: DelayType = 0):
        old_read_delay = self.read_delay
        old_write_delay = self.write_delay

        self.set_delay(read_delay, write_delay)

        try:
            yield
        finally:
            self.set_delay(old_read_delay, old_write_delay)


@pytest.fixture(scope="session")
def tcp_proxy() -> Type[TCPProxy]:
    return TCPProxy


def isasyncgenerator(func):
    if isasyncgenfunction(func):
        return True
    elif asyncio.iscoroutinefunction(func):
        return False


def pytest_configure(config):
    config.addinivalue_line(
        "markers",
        "forbid_get_event_loop: "
        "fail when asyncio.get_event_loop will be called",
    )
    config.addinivalue_line(
        "markers",
        "catch_loop_exceptions: "
        "fails when unhandled loop exception "
        "will be raised",
    )


def pytest_addoption(parser):
    group = parser.getgroup("aiomisc plugin options")
    group.addoption(
        "--aiomisc", action="store_true", default=True,
        help="Use aiomisc entrypoint for run async tests",
    )

    group.addoption(
        "--aiomisc-debug", action="store_true", default=False,
        help="Set debug for event loop",
    )

    group.addoption(
        "--aiomisc-pool-size", type=int, default=4,
        help="Default thread pool size",
    )

    group.addoption(
        "--aiomisc-test-timeout", type=float, default=None,
        help="Test timeout",
    )


def pytest_fixture_setup(fixturedef, request):  # type: ignore
    func = fixturedef.func

    is_async_gen = isasyncgenerator(func)

    if is_async_gen is None:
        return

    strip_request = False
    if "request" not in fixturedef.argnames:
        fixturedef.argnames += ("request",)
        strip_request = True

    if "loop" not in request.fixturenames:
        raise Exception("`loop` fixture required")

    # noinspection PyProtectedMember
    loop_fixturedef = request._get_active_fixturedef("loop")

    def wrapper(*args, **kwargs):  # type: ignore
        if strip_request:
            request = kwargs.pop("request")
        else:
            request = kwargs["request"]

        event_loop = request.getfixturevalue("loop")

        if not is_async_gen:
            return event_loop.run_until_complete(func(*args, **kwargs))

        gen = func(*args, **kwargs)

        def finalizer():  # type: ignore
            try:
                return event_loop.run_until_complete(gen.__anext__())
            except StopAsyncIteration:  # NOQA
                pass

        loop_fixturedef.addfinalizer(
            partial(fixturedef.finish, request=request),
        )

        request.addfinalizer(finalizer)
        return event_loop.run_until_complete(gen.__anext__())

    fixturedef.func = wrapper


@pytest.fixture(scope="session")
def localhost():
    params = (
        (socket.AF_INET, "127.0.0.1"),
        (socket.AF_INET6, "::1"),
    )
    for family, addr in params:
        with socket.socket(family, socket.SOCK_STREAM) as sock:
            try:
                sock.bind((addr, 0))
            except Exception:
                pass
            else:
                return addr
    raise RuntimeError("localhost unavailable")


@pytest.fixture(autouse=True)
def loop_debug(request):
    return request.config.getoption("--aiomisc-debug")


@pytest.fixture(autouse=True)
def aiomisc_test_timeout(request):
    return request.config.getoption("--aiomisc-test-timeout")


@pytest.fixture(autouse=True)
def aiomisc_func_wrap():
    return aiomisc.awaitable


def pytest_pycollect_makeitem(collector, name, obj):  # type: ignore
    if collector.funcnamefilter(name) and asyncio.iscoroutinefunction(obj):
        return list(collector._genfunctions(name, obj))


@pytest.mark.tryfirst
def pytest_pyfunc_call(pyfuncitem):  # type: ignore
    if not asyncio.iscoroutinefunction(pyfuncitem.function):
        return

    event_loop = pyfuncitem.funcargs.get("loop", None)
    func_wraper = pyfuncitem.funcargs.get(
        "aiomisc_func_wrap", aiomisc.awaitable,
    )

    aiomisc_test_timeout = pyfuncitem.funcargs.get(
        "aiomisc_test_timeout", None,
    )

    kwargs = {
        arg: pyfuncitem.funcargs[arg]
        for arg in pyfuncitem._fixtureinfo.argnames
    }

    @wraps(pyfuncitem.obj)
    async def func():
        return await asyncio.wait_for(
            func_wraper(pyfuncitem.obj)(**kwargs),
            timeout=aiomisc_test_timeout,
        )

    event_loop.run_until_complete(func())

    return True


@pytest.fixture(scope="session")
def thread_pool_size(request):
    return request.config.getoption("--aiomisc-pool-size")


@pytest.fixture
def services():
    return []


@pytest.fixture
def default_context():
    return {}


loop_autouse = os.getenv("AIOMISC_LOOP_AUTOUSE", "1") == "1"


@pytest.fixture
def thread_pool_executor():
    from aiomisc.thread_pool import ThreadPoolExecutor
    return ThreadPoolExecutor


@pytest.fixture(autouse=loop_autouse)
def event_loop_policy():
    if uvloop:
        return uvloop.EventLoopPolicy()
    return asyncio.DefaultEventLoopPolicy()


@pytest.fixture
def entrypoint_kwargs() -> dict:
    return {"log_config": False}


@pytest.fixture(name="loop", autouse=loop_autouse)
def _loop(event_loop_policy, caplog):
    basic_config(
        log_format="plain",
        stream=caplog.handler.stream,
    )

    try:
        asyncio.set_event_loop_policy(event_loop_policy)
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        try:
            yield loop
        finally:
            basic_config(
                log_format="plain",
                stream=sys.stderr,
            )

            if loop.is_closed():
                return

            with suppress(Exception):
                loop.run_until_complete(loop.shutdown_asyncgens())
            with suppress(Exception):
                loop.close()
    finally:
        asyncio.set_event_loop_policy(asyncio.DefaultEventLoopPolicy())


@pytest.fixture(autouse=loop_autouse)
def loop(
    request, services, loop_debug, default_context, entrypoint_kwargs,
    thread_pool_size, thread_pool_executor,
    loop: asyncio.AbstractEventLoop, caplog: pytest.LogCaptureFixture,
):
    from aiomisc.context import get_context
    from aiomisc.entrypoint import entrypoint

    if LOG_LEVEL:
        LOG_LEVEL.set(logging.getLogger().getEffectiveLevel())

    pool = thread_pool_executor(thread_pool_size)
    loop.set_default_executor(pool)

    get_marker = request.node.get_closest_marker
    forbid_loop_getter_marker = get_marker("forbid_get_event_loop")
    catch_unhandled_marker = get_marker("catch_loop_exceptions")

    exceptions = list()
    if catch_unhandled_marker:
        loop.set_exception_handler(lambda l, c: exceptions.append(c))

    try:
        with entrypoint(*services, loop=loop, **entrypoint_kwargs):

            ctx = get_context(loop)

            for key, value in default_context.items():
                ctx[key] = value

            if forbid_loop_getter_marker:
                asyncio.get_event_loop.side_effect = partial(
                    pytest.fail, "get_event_loop is forbidden",
                )

            yield loop

            if exceptions:
                logging.error(
                    "Unhandled exceptions found:\n\n\t%s",
                    "\n\t".join(
                        (
                            "Message: {m}\n\t"
                            "Future: {f}\n\t"
                            "Exception: {e}"
                        ).format(
                            m=e["message"],
                            f=repr(e.get("future")),
                            e=repr(e.get("exception")),
                        ) for e in exceptions
                    ),
                )
                pytest.fail("Unhandled exceptions found. See logs.")
    finally:
        asyncio.get_event_loop.side_effect = get_event_loop
        del loop


def _unused_port(*args) -> int:
    with socket.socket(*args) as sock:
        sock.bind(("", 0))
        port = sock.getsockname()[1]
    return port


def get_unused_port(*args) -> int:
    warnings.warn(
        (
            "Do not use get_unused_port directly, use fixture "
            "'aiomisc_unused_port_factory'"
        ),
        DeprecationWarning, stacklevel=2,
    )
    return _unused_port(*args)


class PortSocket(NamedTuple):
    port: int
    socket: socket.socket


@pytest.fixture
def aiomisc_socket_factory(request, localhost) -> Callable[..., PortSocket]:
    """ Returns a """
    def factory(*args, **kwargs) -> PortSocket:
        sock = bind_socket(*args, address=localhost, port=0, **kwargs)
        port = sock.getsockname()[1]

        # Close socket after teardown
        request.addfinalizer(sock.close)

        return PortSocket(port=port, socket=sock)
    return factory


class SocketWrapper(abc.ABC):
    address: str
    port: int

    def __init__(self, *args):
        self._socket_args = args
        self.address = ""
        self.port = 0

    @abc.abstractmethod
    def prepare(self, address: str) -> None:
        raise NotImplementedError

    def close(self):
        pass


class SocketWrapperUnix(SocketWrapper):
    socket: socket.socket
    fd: int

    def __init__(self, *args):
        super().__init__(*args)
        self.socket = socket.socket(*args)
        self.fd = -1

    def close(self):
        self.socket.close()
        if self.fd > 0:
            os.close(self.fd)

    def prepare(self, address: str) -> None:
        self.socket.bind((address, 0))
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock_set_reuseport(self.socket, True)
        self.address, self.port = self.socket.getsockname()[:2]
        # detach socket object and save file descriptor
        self.fd = self.socket.detach()


class SocketWrapperWindows(SocketWrapper):
    def prepare(self, address: str) -> None:
        self.address = address
        self.port = _unused_port(*self._socket_args)


if platform.system() == "Windows":
    socket_wrapper = SocketWrapperWindows
else:
    socket_wrapper = SocketWrapperUnix


@pytest.fixture
def aiomisc_unused_port_factory(
    request: pytest.FixtureRequest, localhost: str,
) -> Callable[[], int]:
    def port_factory(*args) -> int:
        wrapper = socket_wrapper(*args)
        wrapper.prepare("::" if ":" in localhost else "0.0.0.0")
        request.addfinalizer(wrapper.close)
        return wrapper.port
    return port_factory


@pytest.fixture
def aiomisc_unused_port(aiomisc_unused_port_factory) -> int:
    return aiomisc_unused_port_factory()



import asyncio
import hashlib
import hmac
import logging
import os
import signal
import socket
import sys
from types import TracebackType
from typing import Any, Optional, Type

from . import SIGNAL, AddressType, PacketTypes
from .protocol import ADDRESS_FAMILY, SocketIOProtocol


def execute(protocol: SocketIOProtocol) -> None:
    pid = os.getpid()
    try:
        packet_type, func, args, kwargs = protocol.receive()
        logging.debug(
            "Worker PID: %d got task %r", pid, func, extra={"pid": pid},
        )
    except ConnectionError:
        raise

    if packet_type != packet_type.REQUEST:
        raise RuntimeError("Unexpected request")

    try:
        protocol.send((
            PacketTypes.RESULT,
            func(*args, **kwargs),
        ))
    except asyncio.CancelledError:
        logging.exception(
            "Request cancelled for %r", func, extra={"pid": pid},
        )
        protocol.send((
            PacketTypes.CANCELLED,
            asyncio.CancelledError,
        ))
    except (ConnectionError, ConnectionResetError):
        logging.warning(
            "IPC connection error, worker PID: %d exiting.", pid,
            extra={"pid": pid},
        )
        raise
    except Exception as e:
        logging.exception("Exception when processing request")
        protocol.send((
            PacketTypes.EXCEPTION, e,
        ))


def on_cancel_signal(*_: Any) -> None:
    raise asyncio.CancelledError


signal.signal(SIGNAL, on_cancel_signal)


def on_exception(
    exc_type: Optional[Type[BaseException]],
    exc_value: Optional[BaseException],
    exc_tb: Optional[TracebackType],
) -> None:
    if exc_type is None or exc_value is None or exc_tb is None:
        return
    pid = os.getpid()
    logging.exception(
        "Unhandled exception in worker PID: %d",
        pid, exc_info=(exc_type, exc_value, exc_tb), extra={"pid": pid},
    )


def worker(address: AddressType, cookie: bytes, worker_id: bytes) -> None:
    sys.excepthook = on_exception

    with socket.socket(ADDRESS_FAMILY, socket.SOCK_STREAM) as sock:
        try:
            sock.connect(address)
            proto = SocketIOProtocol(sock)
            proto.send((
                PacketTypes.AUTH,
                worker_id,
                hmac.HMAC(
                    cookie, worker_id, digestmod=hashlib.sha256,
                ).digest(),
                os.getpid(),
            ))

            response = proto.receive()

            if response != PacketTypes.AUTH_OK:
                raise RuntimeError(
                    f"Failed to authorize worker on {address!r}",
                )

            while True:
                execute(proto)
        except ConnectionError:
            return


def bad_initializer(
    address: AddressType, cookie: bytes, worker_id: bytes, exc: BaseException,
) -> None:
    with socket.socket(ADDRESS_FAMILY, socket.SOCK_STREAM) as sock:
        sock.connect(address)
        proto = SocketIOProtocol(sock)
        proto.send((
            PacketTypes.BAD_INITIALIZER,
            worker_id,
            hmac.HMAC(cookie, worker_id, digestmod=hashlib.sha256).digest(),
            os.getpid(),
        ))
        proto.send((PacketTypes.EXCEPTION, exc))



import asyncio
import logging
import socket
from contextvars import ContextVar
from typing import Optional


log = logging.getLogger(__name__)

try:
    from time import time_ns
except ImportError:
    from time import time

    def time_ns() -> int:
        return int(time() * 1000000000)


class EventLoopMixin:
    __slots__ = "_loop",

    _loop: Optional[asyncio.AbstractEventLoop]

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        if not getattr(self, "_loop", None):
            self._loop = asyncio.get_running_loop()
        return self._loop   # type: ignore


try:
    import uvloop  # type: ignore
    event_loop_policy = uvloop.EventLoopPolicy()
except ImportError:
    event_loop_policy = asyncio.DefaultEventLoopPolicy()


if hasattr(socket, "TCP_NODELAY"):
    def sock_set_nodelay(sock: socket.socket) -> None:
        if sock.proto != socket.IPPROTO_TCP:
            return None
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
else:
    def sock_set_nodelay(sock: socket.socket) -> None:
        return None

if hasattr(socket, "SO_REUSEPORT"):
    def sock_set_reuseport(sock: socket.socket, reuse_port: bool) -> None:
        sock.setsockopt(
            socket.SOL_SOCKET, socket.SO_REUSEPORT, reuse_port.real,
        )
else:
    def sock_set_reuseport(sock: socket.socket, reuse_port: bool) -> None:
        log.debug(
            "SO_REUSEPORT is not implemented by "
            "underlying library. Skipping.",
        )


EVENT_LOOP: ContextVar = ContextVar("EVENT_LOOP")


def get_current_loop() -> asyncio.AbstractEventLoop:
    loop: Optional[asyncio.AbstractEventLoop] = EVENT_LOOP.get(None)
    if loop is None:
        raise RuntimeError("no current event loop is set")
    return loop


def set_current_loop(loop: asyncio.AbstractEventLoop) -> None:
    EVENT_LOOP.set(loop)


__all__ = (
    "EventLoopMixin",
    "event_loop_policy",
    "get_current_loop",
    "set_current_loop",
    "sock_set_nodelay",
    "sock_set_reuseport",
    "time_ns",
)
